"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const babylon_1 = require("babylon");
const resolveExportedComponent_1 = require("./resolveExportedComponent");
const babel_traverse_1 = require("babel-traverse");
const extractPropTypeCode_1 = require("./extractPropTypeCode");
const opreateGlobal_1 = require("../utils/opreateGlobal");
const t = require("babel-types");
const getJsonObjComments_1 = require("../utils/getJsonObjComments");
const vm = require("vm");
const path = require("path");
const resolveDefaultConfig_1 = require("../utils/resolveDefaultConfig");
const defaultConfig_1 = require("./defaultConfig");
const comments_1 = require("./comments");
const exportPropTypes = require("../plugins/exportPropTypes");
const AST_PARSE_CONFIG = {
    sourceType: 'module',
    plugins: [
        'classProperties',
        'objectRestSpread',
        'throwExpressions',
        'optionalCatchBinding',
        'importMeta',
        'asyncGenerators',
        'exportNamespaceFrom',
        'exportDefaultFrom',
        'classPrivateMethods',
        'classPrivateProperties',
        'decorators',
        'doExpressions'
    ]
};
/**
 * 解析出props依赖, 遵循react-docgen格式
 * @param file 文件地址
 */
function default_1(file, config) {
    config = config || {};
    config = resolveDefaultConfig_1.default(defaultConfig_1.default, config);
    let classDeclarationPath = null;
    const content = fs_1.readFileSync(file, 'utf8');
    const ast = babylon_1.parse(content, AST_PARSE_CONFIG);
    babel_traverse_1.default(ast, {
        Program(path) {
            classDeclarationPath = resolveExportedComponent_1.default(path);
        }
    });
    if (!classDeclarationPath) {
        throw new Error(`cannot find exported react Component in ${file}`);
    }
    const propsTypesPath = getPropTypesPath(classDeclarationPath);
    const defaultPropsPath = getDefaultPropsPath(classDeclarationPath);
    if (!propsTypesPath) {
        throw new Error(`cannot find proptypes defination in ${file}`);
    }
    const value = propsTypesPath.get('value');
    const code = extractPropTypeCode_1.default(propsTypesPath, defaultPropsPath, path.dirname(file), config);
    comments_1.addComments(getJsonObjComments_1.default(value));
    const propTypes = execExtractCode(code, file, config);
    for (let key of Object.keys(propTypes)) {
        if (comments_1.default[key] && propTypes[key]) {
            propTypes[key].description = comments_1.default[key];
        }
    }
    comments_1.clearComments();
    return propTypes;
}
exports.default = default_1;
/**
 * 找到propTypes path
 * @param path reactComponent class声明所在的path
 */
function getPropTypesPath(path) {
    let propTypesPath = null;
    path.traverse({
        ClassProperty: function (path) {
            const node = path.node;
            if (t.isIdentifier(node.key)
                && node.key.name === 'propTypes') {
                propTypesPath = path;
            }
        }
    });
    return propTypesPath;
}
// get path of defaultProps
function getDefaultPropsPath(path) {
    let defaultProps = null;
    path.traverse({
        ClassProperty: function (path) {
            const node = path.node;
            if (t.isIdentifier(node.key)
                && node.key.name === 'defaultProps') {
                defaultProps = path;
            }
        }
    });
    return defaultProps;
}
/**
 * 执行代码, 拿到propTypes
 * @param code 代码内容为字符串
 */
function execExtractCode(code, file, config) {
    let propTypes = null;
    let defaultProps = null;
    const script = new vm.Script(code);
    const sandbox = { global, require: require, console, module, exports, __filename: file, __dirname: path.dirname(file), callback: ((_propTypes, _defaultProps) => {
            propTypes = _propTypes;
            defaultProps = _defaultProps;
        }) };
    config.globalObject['__babelConfig__'] = generateBabelConfig(config);
    opreateGlobal_1.assignToGlobal(config.globalObject);
    vm.createContext(sandbox);
    script.runInContext(sandbox);
    opreateGlobal_1.removeFromGlobal(config.globalObject);
    if (typeof defaultProps === 'object') {
        for (let key of Object.keys(defaultProps)) {
            if (propTypes[key]) {
                propTypes[key].defaultValue = defaultProps[key];
            }
        }
    }
    return propTypes;
}
function generateBabelConfig(config) {
    const exportPropTypesPlugin = [
        exportPropTypes,
        {
            alias: config.alias,
            resolveModule: config.resolveModule
        }
    ];
    const babelConfig = config.babelConfig;
    babelConfig.plugins.push(exportPropTypesPlugin);
    return babelConfig;
}
